import fcsparser
import os
import numpy as np
import pandas as pd
from .models.dataset import get_data, save_csv
from MulticoreTSNE import MulticoreTSNE as TSNE
from umap import UMAP
from umato import UMATO

if __name__ == "__main__":

    seed = 3
    rs = np.random.RandomState(seed)
    percents = [1, 2, 5, 10, 20, 30, 50, 60, 80, 100]
    # percents = [100]
    x, label = get_data('flow')

    for percent in percents:
        p = int(x.shape[0] * (percent / 100))
        print(f"run for {percent} / 100")
        ix = rs.choice(x.shape[0], p, replace=False)
        x_subset = x[ix]
        label_subset = label[ix]

        # save raw csv for atsne
        save_csv('./', alg_name=f"raw_{percent}", data=x_subset, label=label_subset)

        # # load raw csv
        # x_subset = pd.read_csv(f"./raw_{percent}.csv")
        # label_subset = x_subset['label']
        # x_subset = x_subset.drop(columns=['label'])

        # run tsne
        y = TSNE(n_components=2, n_iter=1500, n_jobs=40, random_state=seed, verbose=2).fit_transform(x_subset)
        save_csv('./', alg_name=f"tsne_{percent}", data=y, label=label_subset)

        # # run umap
        # y = UMAP(n_components=2, verbose=True, random_state=seed).fit_transform(x_subset)
        # save_csv('./', alg_name=f"umap_{percent}", data=y, label=label_subset)

        # # run umato
        # y = UMATO(verbose=True, hub_num=300, random_state=seed, ll=label_subset).fit_transform(x_subset)
        # save_csv('./', alg_name=f"umato_{percent}", data=y, label=label_subset)


from .models.dataset import read_data, get_embed_data, get_data
from .utils import GlobalMeasure
import argparse
import glob
import os


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="quantitative comparison of the embedding result"
    )
    parser.add_argument(
        "--algo",
        type=str,
        help="choose algorithm: pca, tsne, umap, topoae, umato",
        default="all",
    )
    parser.add_argument(
        "--data",
        type=str,
        help="choose dataset: spheres, mnist, fmnist, cifar10",
        default="spheres",
    )
    args = parser.parse_args()

    folder_path = f'./visualization/public/results/{args.data}'
    for filename in glob.glob(os.path.join(folder_path, f'{args.algo}*.csv')):
        with open(filename, 'r') as f:


            filenamelist = filename.split("/")
            filetarget = filenamelist[-1]
            c = filetarget.split(".")
            cc = c[0] + '.' +  c[1]
            if args.algo == "isomap":
                cc = c[0]

            # read data & embedding result
            z = get_embed_data(args.data, cc)
            x, label = get_data(args.data)

            gmeasure = GlobalMeasure(x, z)
            dtmkl01_val = gmeasure.dtm_kl(sigma=0.1)
            print(f"{cc}\tDTM_KL01\t{dtmkl01_val}")


#source: https://github.com/ZJULearning/AtSNE/blob/master/tools/txt_to_fvecs.py

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Convert txt file to fvecs/ivecs file.
fvecs/ivecs vector file formats: http://corpus-texmex.irisa.fr/
"""
import argparse
import struct


def main():
    parser = argparse.ArgumentParser(description='Convert txt format file to fvecs')
    parser.add_argument('--data', type=str, help='data name')
    parser.add_argument('--ignore_firstline', action='store_true')
    parser.add_argument('--label', action='store_true')
    parser.add_argument('--dim', type=int, help='data dimension', default=0)
    args = parser.parse_args()

    dim = args.dim
    with open(f'{args.data}.txt', 'r') as input_file:
        with open(f'{args.data}.fvecs', 'wb') as output_file:
            if args.label:
                dim = 0
                output_file.write(struct.pack('=I', 0))  # placeholder

                for line in input_file.readlines():
                    if line:
                        dim += 1
                        output_file.write(struct.pack('=I', int(line)))
                output_file.seek(0)
                output_file.write(struct.pack('=I', dim))
            else:
                for index, line in enumerate(input_file.readlines()):
                    if index == 0:
                        sp = line.split()
                        if args.ignore_firstline:
                            if dim == 0 and len(sp) == 2:
                                dim = int(line.split()[1])
                            continue
                        else:
                            if dim == 0:
                                dim = len(sp)
                        if args.dim == 0:
                            print('guess data dimension {}'.format(dim))

                    line = line.split()
                    line = list(map(float, line[:dim]))
                    output_file.write(struct.pack('=I', dim))
                    output_file.write(struct.pack('={}{}'.format(dim, 'f'), *line))


if __name__ == '__main__':
    main()


import numpy as np
import numba
import umato.distances as dist
from umato.utils import (
    tau_rand_int,
    adjacency_matrix,
    clip,
    rdist,
)


def get_CE(P, Y, d_squared, a, b):
    Q = pow(1 + a * d_squared ** b, -1)
    loss = -P * np.log(Q + 0.001) - (1 - P) * np.log(1 - Q + 0.001)
    return loss.sum() / 1e5


def get_DTM(adj_x, adj_z, sigma=0.1):
    density_x = calc_DTM(adj_x, sigma)
    density_z = calc_DTM(adj_z, sigma)
    return np.abs(density_x - density_z).sum()


def calc_DTM(adj, sigma):
    density = np.sum(np.exp(-(adj ** 2) / sigma), axis=-1)
    return density / density.sum()


def optimize_global_layout(
    P,
    Z,
    a,
    b,
    alpha,
    n_epochs,
    verbose=False,
    savefig=False,
    label=None
):

    costs = []

    for i in range(n_epochs):

        d_squared = np.square(adjacency_matrix(Z))
        z_diff = np.expand_dims(Z, axis=1) - np.expand_dims(Z, axis=0)
        d_inverse = np.expand_dims(pow(1 + a * d_squared ** b, -1), axis=2)

        # Q is the normalized distance in low dimensional space 
        Q = pow(0.001 + d_squared, -1)
        np.fill_diagonal(Q, 0)
        Q = np.dot(1 - P, Q)
        Q /= np.sum(Q, axis=1, keepdims=True)
        # Q /= Q.max()

        grad = np.expand_dims(
            2 * a * b * P * (1e-12 + d_squared) ** (b - 1) - 2 * b * Q, axis=2
        )
        dZ = np.sum(grad * z_diff * d_inverse, axis=1)
        Z -= alpha * dZ

        if verbose:
            tmpZ = Z / np.sum(Z, axis=1, keepdims=True)
            cost = get_CE(P, tmpZ, d_squared, a, b)
            # cost = get_DTM(P, tmpZ, sigma=0.1)
            costs.append(cost)
            print(
                f"[INFO] Current loss: {cost:.6f}, @ iteration: {i+1}/{n_epochs}"
            )

        # if savefig:
        #     if i < 10 or i == 30:
        #         from umato.umato_ import plot_tmptmp
        #         plot_tmptmp(data=Z, label=label, name=f"pic1_global{i}")

    return Z


def nn_layout_optimize(
    head_embedding,
    tail_embedding,
    head,
    tail,
    hub_info,
    n_epochs,
    n_vertices,
    epochs_per_sample,
    a,
    b,
    rng_state,
    gamma,
    learning_rate=1.0,
    negative_sample_rate=5.0,
    parallel=False,
    verbose=False,
    k=0,
    label=None,
):

    (_, dim) = head_embedding.shape
    move_other = head_embedding.shape[0] == tail_embedding.shape[0]
    alpha = learning_rate

    epochs_per_negative_sample = epochs_per_sample / negative_sample_rate
    epoch_of_next_negative_sample = epochs_per_negative_sample.copy()
    epoch_of_next_sample = epochs_per_sample.copy()

    optimize_fn = numba.njit(
        _nn_layout_optimize_single_epoch, fastmath=True, parallel=parallel
    )
    for n in range(n_epochs):
        optimize_fn(
            head_embedding,
            tail_embedding,
            head,
            tail,
            hub_info,
            n_vertices,
            epochs_per_sample,
            a,
            b,
            rng_state,
            gamma,
            dim,
            move_other,
            alpha,
            epochs_per_negative_sample,
            epoch_of_next_negative_sample,
            epoch_of_next_sample,
            n,
        )

        alpha = learning_rate * (1.0 - (float(n) / float(n_epochs)))

        # if verbose and n % 10 == 0:
        #     from umato.umato_ import plot_tmptmp

        #     plot_tmptmp(data=head_embedding, label=label, name=f"pic3_{k}_local{n}")
        #     print("\tcompleted ", n, " / ", n_epochs, "epochs")

    return head_embedding


def _nn_layout_optimize_single_epoch(
    head_embedding,
    tail_embedding,
    head,
    tail,
    hub_info,
    n_vertices,
    epochs_per_sample,
    a,
    b,
    rng_state,
    gamma,
    dim,
    move_other,
    alpha,
    epochs_per_negative_sample,
    epoch_of_next_negative_sample,
    epoch_of_next_sample,
    n,
):
    for i in numba.prange(epochs_per_sample.shape[0]):
        if epoch_of_next_sample[i] <= n:
            j = head[i]  # j == source index
            k = tail[i]  # k == target index

            current = head_embedding[j]  # current == source location
            other = tail_embedding[k]  # other == target location

            dist_squared = rdist(current, other)  # get distance between them

            if dist_squared > 0.0:
                grad_coeff = -2.0 * a * b * pow(dist_squared, b - 1.0)
                grad_coeff /= a * pow(dist_squared, b) + 1.0
            else:
                grad_coeff = 0.0

            for d in range(dim):
                grad_d = clip(grad_coeff * (current[d] - other[d]), 10.0)

                grad_other = 1.0
                if hub_info[k] == 2:
                    grad_other = gamma

                current[d] += grad_d * alpha

                if move_other:
                    # other[d] += -grad_d * alpha
                    other[d] += -grad_d * alpha * grad_other

            epoch_of_next_sample[i] += epochs_per_sample[i]

            n_neg_samples = int(
                (n - epoch_of_next_negative_sample[i]) / epochs_per_negative_sample[i]
            )

            for p in range(n_neg_samples):
                while True:
                    k = tau_rand_int(rng_state) % n_vertices
                    if hub_info[k] > 0:
                        break

                other = tail_embedding[k]
                dist_squared = rdist(current, other)

                if dist_squared > 0.0:
                    grad_coeff = 2.0 * b
                    grad_coeff /= (0.001 + dist_squared) * (
                        a * pow(dist_squared, b) + 1
                    )
                elif j == k:
                    continue
                else:
                    grad_coeff = 0.0

                for d in range(dim):
                    if grad_coeff > 0.0:
                        grad_d = clip(grad_coeff * (current[d] - other[d]), 10.0)
                    else:
                        grad_d = 10.0

                    # current[d] += grad_d * alpha
                    current[d] += grad_d * alpha * gamma

            epoch_of_next_negative_sample[i] += (
                n_neg_samples * epochs_per_negative_sample[i]
            )



'''
This convert data from txt to csv
'''

import argparse
import csv
import pandas as pd

parser = argparse.ArgumentParser(
    description="data name"
)
parser.add_argument(
    "--data",
    type=str,   
    help="choose dataset: spheres, mnist, fmnist, cifar10",
    default="spheres",
)
args = parser.parse_args()


if __name__ == "__main__":
    # lrs = [0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1]
    # perps = [5, 10, 15, 20, 25, 30, 35, 40, 45, 50]

    # # change txt to csv
    # for lr in lrs:
    #     for perp in perps:
    #         with open(f'{args.data}_result_{lr}_{perp}.txt', 'r') as in_file:
    #             stripped = (line.strip() for line in in_file)
    #             lines = (line.split("\t") for line in stripped if line)
    #             with open(f'{args.data}_{lr}_{perp}.csv', 'w') as out_file:
    #                 writer = csv.writer(out_file)
    #                 writer.writerow(('0', '1'))
    #                 writer.writerows(lines)

    # # add label
    # for lr in lrs:
    #     for perp in perps:
    #         x = pd.read_csv(f'{args.data}_{lr}_{perp}.csv', float_precision='round_trip')
    #         df = pd.DataFrame(x)
    #         y = pd.read_csv(f'visualization/public/results/{args.data}/pca.csv', float_precision='round_trip')
    #         df2 = pd.DataFrame(y)
    #         df['label'] = df2['label']

    #         df.to_csv(f"atsne_{args.data}_{lr}_{perp}.csv", index=False)

    samples = [1,2,5,10,20,30,50,60,80,100]

    # change txt to csv
    for sample in samples:
        with open(f'atsne_{sample}.txt', 'r') as in_file:
            stripped = (line.strip() for line in in_file)
            lines = (line.split("\t") for line in stripped if line)
            with open(f'atsne_tmp{sample}.csv', 'w') as out_file:
                writer = csv.writer(out_file)
                writer.writerow(('0', '1'))
                writer.writerows(lines)

    for sample in samples:
        x = pd.read_csv(f'atsne_tmp{sample}.csv', float_precision='round_trip')
        df = pd.DataFrame(x)
        y = pd.read_csv(f'tsne_{sample}.csv', float_precision='round_trip')
        df2 = pd.DataFrame(y)
        df['label'] = df2['label']

        df.to_csv(f"atsne_{sample}.csv", index=False)



'''
This convert data into txt (preparation for fvecs)
'''

import numpy as np
import argparse
from .models.dataset import get_data
# import pandas as pd

parser = argparse.ArgumentParser(
    description="data name"
)
parser.add_argument(
    "--data",
    type=str,
    help="choose dataset: spheres, mnist, fmnist, cifar10, flow",
    default="spheres",
)
args = parser.parse_args()


if __name__ == "__main__":
    # data, _ = get_data(args.data)
    # np.savetxt(f"./{args.data}.txt", data)
    # print(f"{args.data}: {data.shape} with type {type(data)} has been successfully saved!")

    import pandas as pd

    for i in [1,2,5,10,20,30,50,60,80,100]:
        data = pd.read_csv(f'./{args.data}_{i}.csv')
        data = data.drop(columns=['label'])
        np.savetxt(f"./{args.data}_{i}.txt", data)
        print(f"{args.data}_{i}: {data.shape} with type {type(data)} has been successfully saved!")


'''
This will include functions to visualize embedding result 
'''

import numpy as np
from .models.dataset import get_embed_data, save_csv, get_data

def plot_tmptmp(data, label, name):
    import matplotlib.pyplot as plt

    plt.scatter(data[:, 0], data[:, 1], s=2.0, c=label, cmap="Spectral", alpha=1.0)
    cbar = plt.colorbar(boundaries=np.arange(11) - 0.5)
    cbar.set_ticks(np.arange(11))
    plt.title("Embedded")
    plt.savefig(f"./tmp/{name}.png")
    plt.close()


if __name__ == "__main__":
    _, label = get_data('kmnist')
    x = get_embed_data('kmnist', 'topoae')  # spheres, mnist, fmnist, cifar10
    plot_tmptmp(x, label, 'zzz')


"""
This will compare embedding result quantitatively using functions in utils.py
"""

from .models.dataset import read_data
from .utils import GlobalMeasure, LocalMeasure
import pandas as pd
import argparse
import numpy as np

MEASURE_GLOBAL_LIST = [
    # "RMSE",
    # "Kruskal",
    # "Sammon",
    "DTM",
    "DTM_KL1",
    "DTM_KL01",
    "DTM_KL001",
]

MEASURE_LOCAL_LIST = [
    # "Spearman",
    "Trustworthiness",
    "Continuity",
    "MRRE_XZ",
    "MRRE_ZX",
]

ALGO_LIST = ["pca", "isomap", "tsne", "umap", "topoae", "atsne", "umato"]
DATA_LIST = ["spheres"]


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="quantitative comparison of the embedding result"
    )
    parser.add_argument(
        "--algo",
        type=str,
        help="choose algorithm: pca, tsne, umap, topoae, atsne, umato",
        default="all",
    )
    parser.add_argument(
        "--data",
        type=str,
        help="choose dataset: spheres, mnist, fmnist, cifar10",
        default="spheres",
    )
    parser.add_argument(
        "--measure", type=str, help="choose measures: all, global, local", default="all"
    )
    parser.add_argument(
        "--k", type=int, help="number of nearest neighbors", default=5
    )
    parser.add_argument(
        "--load", type=bool, help="load hubs", default=False
    )
    args = parser.parse_args()

    measures = []
    algorithms = []
    values = []

    if args.algo != "all":
        ALGO_LIST = [args.algo]

    for alg in ALGO_LIST:

        print(f"[INFO] Test on [{args.data}] dataset using [{alg}]")

        # read data & embedding result
        x, z, label = read_data(args.data, alg)

        if args.load:
            with open('./hubs.npy', 'rb') as f:
                print("loading hubs")
                hubs = np.load(f)
                x = x[hubs]
                z = z[hubs]
                print(f"xlen: {len(x)}")

        if args.measure == "all" or args.measure == "global":

            gmeasure = GlobalMeasure(x, z)

            algorithms.extend([alg] * len(MEASURE_GLOBAL_LIST))
            measures.extend(MEASURE_GLOBAL_LIST)

            # rmse_val = gmeasure.rmse()
            # kruskal_val = gmeasure.kruskal_stress_measure()
            # sammon_val = gmeasure.sammon_stress()
            dtm_val = gmeasure.dtm()
            dtmkl1_val = gmeasure.dtm_kl(sigma=1.0)
            dtmkl01_val = gmeasure.dtm_kl(sigma=0.1)
            dtmkl001_val = gmeasure.dtm_kl(sigma=0.01)
            values.extend(
                [
                    # rmse_val,
                    # kruskal_val,
                    # sammon_val,
                    dtm_val,
                    dtmkl1_val,
                    dtmkl01_val,
                    dtmkl001_val,
                ]
            )

        if args.measure == "all" or args.measure == "local":

            lmeasure = LocalMeasure(x, z, k=args.k)

            algorithms.extend([alg] * len(MEASURE_LOCAL_LIST))
            measures.extend(MEASURE_LOCAL_LIST)

            # spearman_val = lmeasure.spearmans_rho()
            trust_val = lmeasure.trustworthiness()
            conti_val = lmeasure.continuity()
            mrre_xz_val = lmeasure.mrre_xz()
            mrre_zx_val = lmeasure.mrre_zx()

            values.extend(
                # [spearman_val, trust_val, conti_val, mrre_xz_val, mrre_zx_val]
                [trust_val, conti_val, mrre_xz_val, mrre_zx_val]
            )

        result = pd.DataFrame(
            {"measure": measures, "algorithm": algorithms, "values": values}
        )
        result = result.pivot(index="measure", columns="algorithm", values="values").fillna(
            "NA"
        )
        print(f"{result}\n")



from scipy.spatial import procrustes
import numpy as np
import pandas as pd
import os

if __name__ == "__main__":

    datasets = ['tsne', 'umap', 'atsne', 'umato']
    datasets = ['atsne']
    samples = [1, 2, 5, 10, 20, 30, 50, 60, 80, 100]

    for dataset in datasets:
        print(f"run algorithm: {dataset}")
        for sample in samples:
            path = os.path.join(os.getcwd(), "visualization", "public", "results", "stability")

            d1= pd.read_csv(os.path.join(path, f'{dataset}_{sample}.csv'))
            d2= pd.read_csv(os.path.join(path, f'{dataset}_all.csv'))

            d1_label = d1['label']
            d1_no = d1.drop(columns=['label'])
            d1_no = np.array(d1_no)

            d2 = d2.sort_values(['label'], ascending=True)
            d2_no = d2.drop(columns=['label'])
            d2_no = np.array(d2_no)

            num = len(d1_no)
            ix = np.arange(num)

            _, _, disparity = procrustes(d1_no, d2_no[d1_label])

            # print(f"Disparity {sample}/100 % = {disparity}")
            print(disparity)


# Author: Leland McInnes <leland.mcinnes@gmail.com>
#
# License: BSD 3 clause

from __future__ import print_function

import locale
from warnings import warn

from scipy.optimize import curve_fit

import numpy as np
import scipy.sparse
import numba

import umato.distances as dist
import umato.sparse as sparse
import umato.sparse_nndescent as sparse_nn

from umato.utils import (
    ts,
    fast_knn_indices,
)
from umato.nndescent import (
    nn_descent,
    # initialized_nnd_search,
    # initialise_search,
)
from umato.rp_tree import rptree_leaf_array, make_forest

try:
    # Use pynndescent, if installed (python 3 only)
    from pynndescent import NNDescent
    from pynndescent.distances import named_distances as pynn_named_distances
    from pynndescent.sparse import sparse_named_distances as pynn_sparse_named_distances

    _HAVE_PYNNDESCENT = True
except ImportError:
    _HAVE_PYNNDESCENT = False

locale.setlocale(locale.LC_NUMERIC, "C")

INT32_MIN = np.iinfo(np.int32).min + 1
INT32_MAX = np.iinfo(np.int32).max - 1

SMOOTH_K_TOLERANCE = 1e-5
MIN_K_DIST_SCALE = 1e-3
NPY_INFINITY = np.inf


@numba.njit(
    locals={
        "psum": numba.types.float32,
        "lo": numba.types.float32,
        "mid": numba.types.float32,
        "hi": numba.types.float32,
    },
    fastmath=True,
)  # benchmarking `parallel=True` shows it to *decrease* performance
def smooth_knn_dist(distances, k, n_iter=64, local_connectivity=1.0, bandwidth=1.0):
    """Compute a continuous version of the distance to the kth nearest
    neighbor. That is, this is similar to knn-distance but allows continuous
    k values rather than requiring an integral k. In essence we are simply
    computing the distance such that the cardinality of fuzzy set we generate
    is k.

    Parameters
    ----------
    distances: array of shape (n_samples, n_neighbors)
        Distances to nearest neighbors for each samples. Each row should be a
        sorted list of distances to a given samples nearest neighbors.

    k: float
        The number of nearest neighbors to approximate for.

    n_iter: int (optional, default 64)
        We need to binary search for the correct distance value. This is the
        max number of iterations to use in such a search.

    local_connectivity: int (optional, default 1)
        The local connectivity required -- i.e. the number of nearest
        neighbors that should be assumed to be connected at a local level.
        The higher this value the more connected the manifold becomes
        locally. In practice this should be not more than the local intrinsic
        dimension of the manifold.

    bandwidth: float (optional, default 1)
        The target bandwidth of the kernel, larger values will produce
        larger return values.

    Returns
    -------
    knn_dist: array of shape (n_samples,)
        The distance to kth nearest neighbor, as suitably approximated.

    nn_dist: array of shape (n_samples,)
        The distance to the 1st nearest neighbor for each point.
    """
    target = np.log2(k) * bandwidth
    rho = np.zeros(distances.shape[0], dtype=np.float32)
    result = np.zeros(distances.shape[0], dtype=np.float32)

    mean_distances = np.mean(distances)

    for i in range(distances.shape[0]):
        lo = 0.0
        hi = NPY_INFINITY
        mid = 1.0

        # TODO: This is very inefficient, but will do for now. FIXME
        ith_distances = distances[i]
        non_zero_dists = ith_distances[ith_distances > 0.0]
        if non_zero_dists.shape[0] >= local_connectivity:
            index = int(np.floor(local_connectivity))
            interpolation = local_connectivity - index
            if index > 0:
                rho[i] = non_zero_dists[index - 1]
                if interpolation > SMOOTH_K_TOLERANCE:
                    rho[i] += interpolation * (
                        non_zero_dists[index] - non_zero_dists[index - 1]
                    )
            else:
                rho[i] = interpolation * non_zero_dists[0]
        elif non_zero_dists.shape[0] > 0:
            rho[i] = np.max(non_zero_dists)

        for n in range(n_iter):

            psum = 0.0
            for j in range(1, distances.shape[1]):
                d = distances[i, j] - rho[i]
                if d > 0:
                    psum += np.exp(-(d / mid))
                else:
                    psum += 1.0

            if np.fabs(psum - target) < SMOOTH_K_TOLERANCE:
                break

            if psum > target:
                hi = mid
                mid = (lo + hi) / 2.0
            else:
                lo = mid
                if hi == NPY_INFINITY:
                    mid *= 2
                else:
                    mid = (lo + hi) / 2.0

        result[i] = mid

        # TODO: This is very inefficient, but will do for now. FIXME
        if rho[i] > 0.0:
            mean_ith_distances = np.mean(ith_distances)
            if result[i] < MIN_K_DIST_SCALE * mean_ith_distances:
                result[i] = MIN_K_DIST_SCALE * mean_ith_distances
        else:
            if result[i] < MIN_K_DIST_SCALE * mean_distances:
                result[i] = MIN_K_DIST_SCALE * mean_distances

    return result, rho


def nearest_neighbors(
    X,
    n_neighbors,
    metric,
    angular,
    random_state,
    low_memory=False,
    use_pynndescent=True,
    verbose=False,
):
    """Compute the ``n_neighbors`` nearest points for each data point in ``X``
    under ``metric``. This may be exact, but more likely is approximated via
    nearest neighbor descent.

    Parameters
    ----------
    X: array of shape (n_samples, n_features)
        The input data to compute the k-neighbor graph of.

    n_neighbors: int
        The number of nearest neighbors to compute for each sample in ``X``.

    metric: string or callable
        The metric to use for the computation.

    metric_kwds: dict
        Any arguments to pass to the metric computation function.

    angular: bool
        Whether to use angular rp trees in NN approximation.

    random_state: np.random state
        The random state to use for approximate NN computations.

    low_memory: bool (optional, default False)
        Whether to pursue lower memory NNdescent.

    verbose: bool (optional, default False)
        Whether to print status data during the computation.

    Returns
    -------
    knn_indices: array of shape (n_samples, n_neighbors)
        The indices on the ``n_neighbors`` closest points in the dataset.

    knn_dists: array of shape (n_samples, n_neighbors)
        The distances to the ``n_neighbors`` closest points in the dataset.

    rp_forest: list of trees
        The random projection forest used for searching (if used, None otherwise)
    """
    if verbose:
        print(ts(), "Finding Nearest Neighbors")

    if metric == "precomputed":
        # Note that this does not support sparse distance matrices yet ...
        # Compute indices of n nearest neighbors
        knn_indices = fast_knn_indices(X, n_neighbors)
        # knn_indices = np.argsort(X)[:, :n_neighbors]
        # Compute the nearest neighbor distances
        #   (equivalent to np.sort(X)[:,:n_neighbors])
        knn_dists = X[np.arange(X.shape[0])[:, None], knn_indices].copy()

        rp_forest = []
    else:
        # TODO: Hacked values for now
        n_trees = 5 + int(round((X.shape[0]) ** 0.5 / 20.0))
        n_iters = max(5, int(round(np.log2(X.shape[0]))))

        if _HAVE_PYNNDESCENT and use_pynndescent:
            nnd = NNDescent(
                X,
                n_neighbors=n_neighbors,
                metric=metric,
                metric_kwds={},
                random_state=random_state,
                n_trees=n_trees,
                n_iters=n_iters,
                max_candidates=60,
                low_memory=low_memory,
                verbose=verbose,
            )
            knn_indices, knn_dists = nnd.neighbor_graph
            rp_forest = nnd
        else:
            # Otherwise fall back to nn descent in umap
            if callable(metric):
                _distance_func = metric
            elif metric in dist.named_distances:
                _distance_func = dist.named_distances[metric]
            else:
                raise ValueError("Metric is neither callable, nor a recognised string")

            rng_state = random_state.randint(INT32_MIN, INT32_MAX, 3).astype(np.int64)

            if scipy.sparse.isspmatrix_csr(X):
                if callable(metric):
                    _distance_func = metric
                else:
                    try:
                        _distance_func = sparse.sparse_named_distances[metric]
                    except KeyError as e:
                        raise ValueError(
                            "Metric {} not supported for sparse data".format(metric)
                        ) from e

                if verbose:
                    print(ts(), "Building RP forest with", str(n_trees), "trees")

                rp_forest = make_forest(X, n_neighbors, n_trees, rng_state, angular)
                leaf_array = rptree_leaf_array(rp_forest)

                if verbose:
                    print(ts(), "NN descent for", str(n_iters), "iterations")
                knn_indices, knn_dists = sparse_nn.sparse_nn_descent(
                    X.indices,
                    X.indptr,
                    X.data,
                    X.shape[0],
                    n_neighbors,
                    rng_state,
                    max_candidates=60,
                    sparse_dist=distance_func,
                    low_memory=low_memory,
                    rp_tree_init=True,
                    leaf_array=leaf_array,
                    n_iters=n_iters,
                    verbose=verbose,
                )
            else:
                # metric_nn_descent = make_nn_descent(
                #     distance_func, tuple(metric_kwds.values())
                # )

                distance_func = _distance_func

                if verbose:
                    print(ts(), "Building RP forest with", str(n_trees), "trees")

                rp_forest = make_forest(X, n_neighbors, n_trees, rng_state, angular)
                leaf_array = rptree_leaf_array(rp_forest)  # stacked rp_tree indices

                if verbose:
                    print(ts(), "NN descent for", str(n_iters), "iterations")
                knn_indices, knn_dists = nn_descent(
                    X,
                    n_neighbors,
                    rng_state,
                    max_candidates=60,
                    dist=distance_func,
                    low_memory=low_memory,
                    rp_tree_init=True,
                    leaf_array=leaf_array,
                    n_iters=n_iters,
                    verbose=verbose,
                )

            if np.any(knn_indices < 0):
                warn(
                    "Failed to correctly find n_neighbors for some samples."
                    "Results may be less than ideal. Try re-running with"
                    "different parameters."
                )
    if verbose:
        print(ts(), "Finished Nearest Neighbor Search")
    return knn_indices, knn_dists, rp_forest


@numba.njit(
    locals={
        "knn_dists": numba.types.float32[:, ::1],
        "sigmas": numba.types.float32[::1],
        "rhos": numba.types.float32[::1],
        "val": numba.types.float32,
    },
    parallel=True,
    fastmath=True,
)
def compute_membership_strengths(knn_indices, knn_dists, sigmas, rhos, hubs):
    """Construct the membership strength data for the 1-skeleton of each local
    fuzzy simplicial set -- this is formed as a sparse matrix where each row is
    a local fuzzy simplicial set, with a membership strength for the
    1-simplex to each other data point.

    Parameters
    ----------
    knn_indices: array of shape (n_samples, n_neighbors)
        The indices on the ``n_neighbors`` closest points in the dataset.

    knn_dists: array of shape (n_samples, n_neighbors)
        The distances to the ``n_neighbors`` closest points in the dataset.

    sigmas: array of shape(n_samples)
        The normalization factor derived from the metric tensor approximation.

    rhos: array of shape(n_samples)
        The local connectivity adjustment.

    Returns
    -------
    rows: array of shape (n_samples * n_neighbors)
        Row data for the resulting sparse matrix (coo format)

    cols: array of shape (n_samples * n_neighbors)
        Column data for the resulting sparse matrix (coo format)

    vals: array of shape (n_samples * n_neighbors)
        Entries for the resulting sparse matrix (coo format)
    """
    n_samples = len(hubs)
    n_neighbors = knn_indices.shape[1]

    rows = np.zeros(n_samples * n_neighbors, dtype=np.int32)
    cols = np.zeros(n_samples * n_neighbors, dtype=np.int32)
    vals = np.zeros(n_samples * n_neighbors, dtype=np.float32)

    for i in range(n_samples):
        for j in range(n_neighbors):
            if knn_indices[i, j] == -1:
                continue  # We didn't get the full knn for i
            if knn_indices[i, j] == i:
                val = 0.0
            elif knn_dists[i, j] - rhos[i] <= 0.0 or sigmas[i] == 0.0:
                val = 1.0
            else:
                val = np.exp(-((knn_dists[i, j] - rhos[i]) / (sigmas[i])))

            rows[i * n_neighbors + j] = hubs[i]
            cols[i * n_neighbors + j] = knn_indices[i, j]
            vals[i * n_neighbors + j] = val

    return rows, cols, vals


def fuzzy_simplicial_set(
    X,
    n_neighbors,
    random_state,
    metric,
    hubs=None,
    knn_indices=None,
    knn_dists=None,
    angular=False,
    set_op_mix_ratio=1.0,
    local_connectivity=1.0,
    apply_set_operations=True,
    verbose=False,
):
    """Given a set of data X, a neighborhood size, and a measure of distance
    compute the fuzzy simplicial set (here represented as a fuzzy graph in
    the form of a sparse matrix) associated to the data. This is done by
    locally approximating geodesic distance at each point, creating a fuzzy
    simplicial set for each such point, and then combining all the local
    fuzzy simplicial sets into a global one via a fuzzy union.

    Parameters
    ----------
    X: array of shape (n_samples, n_features)
        The data to be modelled as a fuzzy simplicial set.

    n_neighbors: int
        The number of neighbors to use to approximate geodesic distance.
        Larger numbers induce more global estimates of the manifold that can
        miss finer detail, while smaller values will focus on fine manifold
        structure to the detriment of the larger picture.

    random_state: numpy RandomState or equivalent
        A state capable being used as a numpy random state.

    metric: string or function (optional, default 'euclidean')
        The metric to use to compute distances in high dimensional space.
        If a string is passed it must match a valid predefined metric. If
        a general metric is required a function that takes two 1d arrays and
        returns a float can be provided. For performance purposes it is
        required that this be a numba jit'd function. Valid string metrics
        include:
            * euclidean (or l2)
            * manhattan (or l1)
            * cityblock
            * braycurtis
            * canberra
            * chebyshev
            * correlation
            * cosine
            * dice
            * hamming
            * jaccard
            * kulsinski
            * ll_dirichlet
            * mahalanobis
            * matching
            * minkowski
            * rogerstanimoto
            * russellrao
            * seuclidean
            * sokalmichener
            * sokalsneath
            * sqeuclidean
            * yule
            * wminkowski

        Metrics that take arguments (such as minkowski, mahalanobis etc.)
        can have arguments passed via the metric_kwds dictionary. At this
        time care must be taken and dictionary elements must be ordered
        appropriately; this will hopefully be fixed in the future.

    knn_indices: array of shape (n_samples, n_neighbors) (optional)
        If the k-nearest neighbors of each point has already been calculated
        you can pass them in here to save computation time. This should be
        an array with the indices of the k-nearest neighbors as a row for
        each data point.

    knn_dists: array of shape (n_samples, n_neighbors) (optional)
        If the k-nearest neighbors of each point has already been calculated
        you can pass them in here to save computation time. This should be
        an array with the distances of the k-nearest neighbors as a row for
        each data point.

    angular: bool (optional, default False)
        Whether to use angular/cosine distance for the random projection
        forest for seeding NN-descent to determine approximate nearest
        neighbors.

    set_op_mix_ratio: float (optional, default 1.0)
        Interpolate between (fuzzy) union and intersection as the set operation
        used to combine local fuzzy simplicial sets to obtain a global fuzzy
        simplicial sets. Both fuzzy set operations use the product t-norm.
        The value of this parameter should be between 0.0 and 1.0; a value of
        1.0 will use a pure fuzzy union, while 0.0 will use a pure fuzzy
        intersection.

    local_connectivity: int (optional, default 1)
        The local connectivity required -- i.e. the number of nearest
        neighbors that should be assumed to be connected at a local level.
        The higher this value the more connected the manifold becomes
        locally. In practice this should be not more than the local intrinsic
        dimension of the manifold.

    verbose: bool (optional, default False)
        Whether to report information on the current progress of the algorithm.

    Returns
    -------
    fuzzy_simplicial_set: coo_matrix
        A fuzzy simplicial set represented as a sparse matrix. The (i,
        j) entry of the matrix represents the membership strength of the
        1-simplex between the ith and jth sample points.
    """
    if knn_indices is None or knn_dists is None:
        knn_indices, knn_dists, _ = nearest_neighbors(
            X, n_neighbors, metric, angular, random_state, verbose=verbose
        )

    knn_dists = knn_dists.astype(np.float32)

    sigmas, rhos = smooth_knn_dist(
        knn_dists, float(n_neighbors), local_connectivity=float(local_connectivity),
    )

    if hubs is not None:  # build graph using only (hub + nn) nodes
        rows, cols, vals = compute_membership_strengths(
            knn_indices, knn_dists, sigmas, rhos, hubs,
        )
    else:
        raise ValueError(f"Check if hubs are computed correctly! hubs={hubs}")

    result = scipy.sparse.coo_matrix(
        (vals, (rows, cols))
    )  # (TODO) do I need to set the shape ?
    result.eliminate_zeros()

    if apply_set_operations:
        transpose = result.transpose()

        prod_matrix = result.multiply(transpose)

        result = (
            set_op_mix_ratio * (result + transpose - prod_matrix)
            + (1.0 - set_op_mix_ratio) * prod_matrix
        )

    result.eliminate_zeros()

    return result, sigmas, rhos


def make_epochs_per_sample(weights, n_epochs):
    """Given a set of weights and number of epochs generate the number of
    epochs per sample for each weight.

    Parameters
    ----------
    weights: array of shape (n_1_simplices)
        The weights ofhow much we wish to sample each 1-simplex.

    n_epochs: int
        The total number of epochs we want to train for.

    Returns
    -------
    An array of number of epochs per sample, one for each 1-simplex.
    """
    result = -1.0 * np.ones(weights.shape[0], dtype=np.float64)
    n_samples = n_epochs * (weights / weights.max())
    result[n_samples > 0] = float(n_epochs) / n_samples[n_samples > 0]
    return result


def find_ab_params(spread, min_dist):
    """Fit a, b params for the differentiable curve used in lower
    dimensional fuzzy simplicial complex construction. We want the
    smooth curve (from a pre-defined family with simple gradient) that
    best matches an offset exponential decay.
    """

    def curve(x, a, b):
        return 1.0 / (1.0 + a * x ** (2 * b))

    xv = np.linspace(0, spread * 3, 300)
    yv = np.zeros(xv.shape)
    yv[xv < min_dist] = 1.0
    yv[xv >= min_dist] = np.exp(-(xv[xv >= min_dist] - min_dist) / spread)
    params, covar = curve_fit(curve, xv, yv)
    return params[0], params[1]



"""
Author: Hyung-Kwon Ko (hkko@hcil.snu.ac.kr)
"""

from __future__ import print_function

import locale
from warnings import warn
import time

from sklearn.base import BaseEstimator
from sklearn.utils import check_random_state, check_array
from sklearn.neighbors import KDTree
from sklearn.decomposition import PCA

try:
    import joblib
except ImportError:
    # sklearn.externals.joblib is deprecated in 0.21, will be removed in 0.23
    from sklearn.externals import joblib

import numpy as np
import scipy.sparse
import scipy.sparse.csgraph
import numba

import umato.distances as dist
import umato.sparse as sparse
from sklearn.manifold import SpectralEmbedding

from umato.utils import (
    adjacency_matrix,
    ts,
    csr_unique,
    plot_tmptmp,
)

from umato.layouts import (
    optimize_global_layout,
    nn_layout_optimize,
)

from umato.umap_ import (
    nearest_neighbors,
    fuzzy_simplicial_set,
    make_epochs_per_sample,
    find_ab_params,
)

try:
    # Use pynndescent, if installed (python 3 only)
    from pynndescent import NNDescent
    from pynndescent.distances import named_distances as pynn_named_distances
    from pynndescent.sparse import sparse_named_distances as pynn_sparse_named_distances

    _HAVE_PYNNDESCENT = True
except ImportError:
    _HAVE_PYNNDESCENT = False


locale.setlocale(locale.LC_NUMERIC, "C")

INT32_MIN = np.iinfo(np.int32).min + 1
INT32_MAX = np.iinfo(np.int32).max - 1


@numba.njit(parallel=True, fastmath=True)
def calculate_distances(data, source, targets):
    distances_squared = np.zeros(len(targets), dtype=data.dtype)
    for target_idx in numba.prange(len(targets)):
        target = targets[target_idx]
        distance_squared = np.sum((data[source] - data[target]) ** 2)
        distances_squared[target_idx] = distance_squared
    return distances_squared


@numba.njit(
    # parallel=True,  # can SABOTAGE the array order (should be used with care)
    fastmath=True,
)
def build_knn_graph(
    data, sorted_index, hub_num,
):
    sorted_index_c = sorted_index.copy()

    index_map = [0] * len(sorted_index_c)
    for i, val in enumerate(sorted_index_c):
        index_map[val] = i
    index_map = np.array(index_map)

    disjoints = []

    real_point_count = data.shape[0]
    # the number of nodes per KNN, i.e. the K value
    leaf_num = int(np.ceil(real_point_count / hub_num))

    # the number of points to be appended to the disjoints list
    total_disjoints_count = leaf_num * hub_num

    # the number of padding points, having index value -1
    # this number will be smaller than leaf_num, as empty KNNs won't be created
    padding_point_count = (total_disjoints_count - real_point_count) % leaf_num

    # the number of hubs with no padding points
    full_real_knn_hub_count = int(np.floor(real_point_count / leaf_num))

    last_source_loc = -1

    # deal with hubs with no padding points first
    for i in range(full_real_knn_hub_count):
        # append the first element
        for j in range(last_source_loc + 1, len(sorted_index_c)):
            if sorted_index_c[j] > -1:
                source = sorted_index_c[j]
                sorted_index_c[j] = -1
                last_source_loc = j
                break
        else:
            # No hub source found!
            raise AssertionError("Logic should not reach this block.")

        # get distance for each element
        targets = sorted_index[sorted_index_c != -1]
        distances_squared = calculate_distances(data, source, targets)

        # append other elements
        sorted_distance_indexes = targets[np.argsort(distances_squared)[:leaf_num - 1]].astype(np.int64)

        # create KNN
        disjoint = [source] + list(sorted_distance_indexes)

        # mark indexes as used
        sorted_index_c[index_map[sorted_distance_indexes]] = -1

        disjoints.append(disjoint)

    if padding_point_count:
        source = -1
        disjoint = []

        # append the first element
        for j in range(last_source_loc + 1, len(sorted_index_c)):
            if sorted_index_c[j] > -1:
                source = sorted_index_c[j]
                disjoint.append(source)
                sorted_index_c[j] = -1
                last_source_loc = j
                break
        if source == -1:
            # No hub source found!
            raise AssertionError("Logic should not reach this block.")

        # get distance for each element
        targets = sorted_index[sorted_index_c != -1]
        distances_squared = calculate_distances(data, source, targets)

        # append other elements
        real_elements = leaf_num - padding_point_count
        sorted_distance_indexes = targets[np.argsort(distances_squared)[:real_elements - 1]].astype(np.int64)

        # create KNN
        disjoint = [source] + list(sorted_distance_indexes) + [-1] * padding_point_count

        # mark indexes as used
        sorted_index_c[index_map[sorted_distance_indexes]] = -1

        disjoints.append(disjoint)

    return np.array(disjoints)


def pick_hubs(
    disjoints, random_state, popular=False,
):
    if popular:
        return disjoints[:, 0]
    else:
        hubs = []
        (hub_num, _) = disjoints.shape

        # append until second to last element
        for i in range(hub_num - 1):
            choice = random_state.choice(disjoints[i])
            hubs.append(choice)

        # append last element
        last = disjoints[hub_num - 1]
        last = last[last != -1]
        choice = random_state.choice(last)
        hubs.append(choice)

        if hub_num != len(hubs):
            ValueError(f"hub_num({hub_num}) is not the same as hubs({hubs})!")

        return hubs


def build_global_structure(
    data,
    hubs,
    n_components,
    a,
    b,
    random_state,
    alpha=0.0065,
    n_epochs=30,
    verbose=False,
    label=None,
    init_global="pca",
):
    if isinstance(init_global, str):
        if init_global == "pca":
            Z = PCA(n_components=n_components).fit_transform(data[hubs])
            Z /= Z.max()
        elif init_global == "random":
            Z = random_state.normal(
                loc=0.0, scale=0.05, size=list((len(hubs), n_components))
            ).astype(np.float32)
            Z /= Z.max()
        elif init_global == "spectral":
            Z = SpectralEmbedding(n_components=n_components).fit_transform(data[hubs])
        else:
            raise ValueError("Check hub node initializing method!")
    else:
        Z = init_global[hubs]
        Z /= Z.max()

    P = adjacency_matrix(data[hubs])
    # P /= np.sum(P, axis=1, keepdims=True)
    P /= P.max()

    if verbose:
        result = optimize_global_layout(
            P=P,
            Z=Z,
            a=a,
            b=b,
            alpha=alpha,
            n_epochs=n_epochs,
            verbose=verbose,
            savefig=verbose,
            label=label[hubs] if label is not None else None,
        )
    else:
        result = optimize_global_layout(
            P, Z, a, b, alpha=alpha, n_epochs=n_epochs
        )  # (TODO) how to optimize n_epochs & alpha?

    return result


def embed_others_nn(
    data, init_global, hubs, knn_indices, nn_consider, random_state, label, verbose=False,
):
    init = np.zeros((data.shape[0], init_global.shape[1]))
    original_hubs = hubs.copy()
    init[original_hubs] = init_global

    while True:
        val = len(hubs)
        hubs = hub_nn_num(
            data=data, hubs=hubs, knn_indices=knn_indices, nn_consider=nn_consider,
        )

        if val == len(hubs):
            if len(init) > len(hubs) and verbose:
                print(f"len(hubs) {len(hubs)} is smaller than len(init) {len(init)}")
            break
		
    # generate random normal distribution
    random_normal = random_state.normal(
        loc=0.0, scale=0.05, size=list(init.shape)
    ).astype(np.float32)

    hub_nn = set(hubs) - set(original_hubs)
    hub_nn = np.array(list(hub_nn))

    # initialize other nodes' position using only hub information
    init = nn_initialize(
        data=data,
        init=init,
        original_hubs=original_hubs,
        hub_nn=hub_nn,
        random=random_normal,
        nn_consider=10 if len(original_hubs) >= 10 else len(original_hubs), # number of hubs to consider
    )

    # np.array of hub information (hubs = 2, hub_nn = 1, outliers = 0)
    hub_info = np.zeros(data.shape[0])
    hub_info[hub_nn] = 1
    hub_info[original_hubs] = 2

    # save figure2
    # if verbose:
    #     plot_tmptmp(data=init[hubs], label=label[hubs] if label is not None else None, name=f"pic2")

    return init, hub_info, hubs


def embed_outliers(
    data, init, hubs, disjoints, random_state, label, n_neighbors, verbose=False,
):
    # generate random normal distribution
    random_normal = random_state.normal(scale=0.02, size=list(init.shape)).astype(
        np.float32
    )

    # append other nodes using NN disjoint information
    init, nodes_number = disjoint_initialize(
        data=data, init=init, hubs=hubs, disjoints=disjoints, random=random_normal, nn_consider=n_neighbors
    )

    if len(init) != len(nodes_number):
        raise ValueError(
            f"total data # ({len(init)}) != total embedded # ({len(nodes_number)})!"
        )

    # save figure3
    # if verbose:
    #     plot_tmptmp(data=init, label=label, name="pic4_disjoint")

    return init


@numba.njit()
def disjoint_initialize(
    data, init, hubs, disjoints, random, nn_consider
):

    hubs_true = np.zeros(data.shape[0])
    hubs_true[hubs] = True
    hubs = set(hubs)

    for disjoint in disjoints:
        for j in disjoint:
            # j == -1 means we've run all the iteration
            if j == -1:
                break
            # if it is not a hub node, we should embed this using NN in disjoint set
            if not hubs_true[j]:
                distances = []
                indices = []
                # we use its neighbors
                for k in disjoint:
                    if hubs_true[k]:
                        distance = 0.0
                        for l in range(data.shape[1]):
                            distance += (data[j][l] - data[k][l]) ** 2
                        distance = np.sqrt(distance)
                        distances.append(distance)
                        indices.append(k)

                nn_consider_tmp = nn_consider
                if len(distances) < nn_consider:
                    nn_consider_tmp = len(distances)

                ixs = np.array(distances).argsort()[:nn_consider_tmp]
                init[j] = np.zeros(init.shape[1])
                for ix in ixs:
                    target_ix = indices[ix]
                    init[j] += init[target_ix]
                init[j] /= nn_consider_tmp

                hubs.add(j)

    return init, hubs


@numba.njit()
def hub_nn_num(
    data, hubs, knn_indices, nn_consider=10,
):
    num_log = np.zeros(data.shape[0])
    num_log[hubs] = -1

    hubs = set(hubs)
    hubs_fin = hubs.copy()

    for i in hubs:
        for j, e in enumerate(knn_indices[i]):
            if j > nn_consider:
                break
            if num_log[e] > -1:
                hubs_fin.add(e)

    return np.array(list(hubs_fin))


@numba.njit(
    locals={
        "num_log": numba.types.float32[::1],
        "index": numba.types.int32,
        "dists": numba.types.float32[::1],
        "dist": numba.types.float32,
    },
    parallel=True,
    fastmath=True,
)
def nn_initialize(
    data, init, original_hubs, hub_nn, random, nn_consider=10,
):

    num_log = np.zeros(data.shape[0], dtype=np.float32)
    num_log[original_hubs] = -1
    num_log[hub_nn] = -1


    for i in numba.prange(len(hub_nn)):
        # find nearest hub nodes
        dists = np.zeros(len(original_hubs), dtype=np.float32)
        for j in numba.prange(len(original_hubs)):
            dist = 0.0
            for d in numba.prange(data.shape[1]):
                e = original_hubs[j]
                dist += (data[e][d] - data[hub_nn[i]][d]) ** 2
            dists[j] = dist

        # sorted hub indices
        dists_arg = dists.argsort(kind="quicksort")

        for k in numba.prange(nn_consider):
            index = original_hubs[dists_arg[k]]
            init[hub_nn[i]] += init[index]
            num_log[hub_nn[i]] += 1

        # add random value before break
        init[hub_nn[i]] += random[hub_nn[i]]

    for l in numba.prange(data.shape[0]):
        if num_log[l] > 0:
            init[l] /= num_log[l]

    return init


@numba.njit(
    locals={
        "out_indices": numba.types.int32[:, ::1],
        "out_dists": numba.types.float32[:, ::1],
        "counts": numba.types.int32[::1],
    },
    parallel=True,
    fastmath=True,
)
def select_from_knn(
    knn_indices, knn_dists, hub_info, n_neighbors, n,
):
    out_indices = np.zeros((n, n_neighbors), dtype=np.int32)
    out_dists = np.zeros((n, n_neighbors), dtype=np.float32)
    counts = np.zeros(n, dtype=np.int32)

    for i in numba.prange(knn_indices.shape[0]):
        if hub_info[i] > 0:
            for j in range(knn_indices.shape[1]):
                # append directly if it is not an outlier
                if hub_info[knn_indices[i, j]] > 0:
                    out_indices[i, counts[i]] = knn_indices[i, j]
                    out_dists[i, counts[i]] = knn_dists[i, j]
                    counts[i] += 1
                if counts[i] == n_neighbors:
                    break

    return out_indices, out_dists, counts


@numba.njit(
    # locals={"dists": numba.types.float32[::1],},
    parallel=True,
    fastmath=True,
)
def apppend_knn(
    data, knn_indices, knn_dists, hub_info, n_neighbors, counts, counts_sum,
):
    for i in numba.prange(data.shape[0]):
        num = n_neighbors - counts[i]
        if hub_info[i] > 0 and num > 0:
            # found neighbors (# of neighbors < n_neighbors)
            neighbors = knn_indices[i][: counts[i]]

            # find unique target indices
            indices = set()
            for ci in range(counts[i]):  # cannot use numba.prange; malloc error occurs
                cx = neighbors[ci]
                for cy in range(counts[cx]):
                    indices.add(knn_indices[cx][cy])

            # get target indices
            targets = indices - set(neighbors)
            targets = np.array(list(targets))

            # if there is not enough target, it is a corner case (raise error)
            if len(targets) < num:
                return knn_indices, knn_dists, -1
            else:
                # calculate distances
                dists = np.zeros(len(targets), dtype=np.float32)
                for k in numba.prange(len(targets)):
                    dist = 0.0
                    for d in numba.prange(data.shape[1]):
                        dist += (data[i][d] - data[targets[k]][d]) ** 2
                    dists[k] = np.sqrt(dist)
                sorted_dists_index = dists.argsort(kind="quicksort")

                # add more knns
                for j in numba.prange(num):
                    knn_indices[i][counts[i] + j] = targets[
                        sorted_dists_index[counts[i] + j]
                    ]
                    knn_dists[i][counts[i] + j] = dists[
                        sorted_dists_index[counts[i] + j]
                    ]

                # re-sort index
                sorted_knn_index = knn_dists[i].argsort(kind="quicksort")
                knn_indices[i] = knn_indices[i][sorted_knn_index]
                knn_dists[i] = knn_dists[i][sorted_knn_index]

                # for double check
                counts_sum -= 1

    return knn_indices, knn_dists, counts_sum


def local_optimize_nn(
    data,
    graph,
    hub_info,
    n_components,
    learning_rate,
    a,
    b,
    gamma,
    negative_sample_rate,
    n_epochs,
    init,
    random_state,
    parallel=False,
    verbose=False,
    label=None,
):

    graph = graph.tocoo()
    graph.sum_duplicates()
    n_vertices = graph.shape[1]

    graph.data[
        hub_info[graph.col] == 2
    ] = 1.0  # current (NNs) -- other (hubs): 1.0 weight
    graph.data[
        hub_info[graph.row] == 2
    ] = 0.0  # current (hubs) -- other (hubs, nns): 0.0 weight (remove)
    graph.data[graph.data < (graph.data.max() / float(n_epochs))] = 0.0
    graph.eliminate_zeros()

    init_data = np.array(init)
    if len(init_data.shape) == 2:
        if np.unique(init_data, axis=0).shape[0] < init_data.shape[0]:
            tree = KDTree(init_data)
            dist, ind = tree.query(init_data, k=2)
            nndist = np.mean(dist[:, 1])
            embedding = init_data + random_state.normal(
                scale=0.001 * nndist, size=init_data.shape
            ).astype(np.float32)
        else:
            embedding = init_data

    epochs_per_sample = make_epochs_per_sample(graph.data, n_epochs)

    head = graph.row
    tail = graph.col

    embedding = (
        10.0
        * (embedding - np.min(embedding, 0))
        / (np.max(embedding, 0) - np.min(embedding, 0))
    ).astype(np.float32, order="C")

    rng_state = random_state.randint(INT32_MIN, INT32_MAX, 3).astype(np.int64)

    embedding = nn_layout_optimize(
        embedding,
        embedding,
        head,
        tail,
        hub_info,
        n_epochs,
        n_vertices,
        epochs_per_sample,
        a,
        b,
        rng_state,
        gamma=gamma,
        learning_rate=learning_rate,
        negative_sample_rate=negative_sample_rate,
        parallel=parallel,
        verbose=verbose,
        label=label,
    )

    return embedding


class UMATO(BaseEstimator):
    def __init__(
        self,
        n_neighbors=50,
        n_components=2,
        hub_num=300,
        metric="euclidean",
        global_n_epochs=None,
        local_n_epochs=None,
        global_learning_rate=0.0065,
        local_learning_rate=0.01,
        min_dist=0.1,
        spread=1.0,
        low_memory=False,
        set_op_mix_ratio=1.0,
        local_connectivity=1.0,
        gamma=0.1,
        negative_sample_rate=5,
        a=None,
        b=None,
        random_state=None,
        angular_rp_forest=False,
        init="pca",
        ll=None,
        verbose=False,
    ):
        self.n_neighbors = n_neighbors
        self.hub_num = hub_num
        self.metric = metric
        self.global_n_epochs = global_n_epochs
        self.local_n_epochs = local_n_epochs
        self.n_components = n_components
        self.gamma = gamma
        self.global_learning_rate = global_learning_rate
        self.local_learning_rate = local_learning_rate
        self.spread = spread
        self.min_dist = min_dist
        self.low_memory = low_memory
        self.set_op_mix_ratio = set_op_mix_ratio
        self.local_connectivity = local_connectivity
        self.negative_sample_rate = negative_sample_rate
        self.random_state = random_state
        self.angular_rp_forest = angular_rp_forest
        self.verbose = verbose
        self.a = a
        self.b = b
        self.init = init

        self.ll = ll

    def _validate_parameters(self):
        if self.hub_num >= self._raw_data.shape[0]:
            raise ValueError("hub_num must be less than the number of data points")
        if self.hub_num < 2:
            raise ValueError("hub_num cannot be smaller than 2")
        if self.set_op_mix_ratio < 0.0 or self.set_op_mix_ratio > 1.0:
            raise ValueError("set_op_mix_ratio must be between 0.0 and 1.0")
        if self.gamma < 0.0:
            raise ValueError("gamma cannot be negative")
        if self.min_dist > self.spread:
            raise ValueError("min_dist must be less than or equal to spread")
        if self.min_dist < 0.0:
            raise ValueError("min_dist cannot be negative")
        if not isinstance(self.metric, str) and not callable(self.metric):
            raise ValueError("metric must be string or callable")
        if self.negative_sample_rate < 0:
            raise ValueError("negative sample rate must be positive")
        if self.global_learning_rate < 0.0 or self.local_learning_rate < 0.0:
            raise ValueError("learning_rates must be positive")
        if self.n_neighbors < 2:
            raise ValueError("n_neighbors must be greater than 1")
        if not isinstance(self.hub_num, int) or self.hub_num < -1:
            raise ValueError("hub_num must be a positive integer or -1 (None)")
        if not isinstance(self.n_components, int):
            if isinstance(self.n_components, str):
                raise ValueError("n_components must be an int")
            if self.n_components % 1 != 0:
                raise ValueError("n_components must be a whole number")
            try:
                # this will convert other types of int (eg. numpy int64)
                # to Python int
                self.n_components = int(self.n_components)
            except ValueError:
                raise ValueError("n_components must be an int")
        if self.n_components < 1:
            raise ValueError("n_components must be greater than 0")

        if self.global_n_epochs is not None and (
            self.global_n_epochs <= 10 or not isinstance(self.global_n_epochs, int)
        ):
            raise ValueError("global_n_epochs must be a positive integer of at least 10")
        if self.local_n_epochs is not None and (
            self.local_n_epochs <= 10 or not isinstance(self.local_n_epochs, int)
        ):
            raise ValueError("local_n_epochs must be a positive integer of at least 10")

        # check sparsity of data
        if scipy.sparse.isspmatrix_csr(self._raw_data):
            self._sparse_data = True
        else:
            self._sparse_data = False

        # set input distance metric
        if self.metric in dist.named_distances:
            if self._sparse_data:
                if self.metric in sparse.sparse_named_distances:
                    self._input_distance_func = sparse.sparse_named_distances[
                        self.metric
                    ]
                else:
                    raise ValueError(
                        "Metric {} is not supported for sparse data".format(self.metric)
                    )
            else:
                self._input_distance_func = dist.named_distances[self.metric]
        else:
            raise ValueError("metric is not a recognised string")

        # set angularity for NN search based on metric
        if self.metric in (
            "cosine",
            "correlation",
            "dice",
            "jaccard",
            "ll_dirichlet",
            "hellinger",
        ):
            self.angular_rp_forest = True

    def fit(self, X):
        """Fit X into an embedded space.

        Optionally use y for supervised dimension reduction.

        Parameters
        ----------
        X : array, shape (n_samples, n_features) or (n_samples, n_samples)
            If the metric is 'precomputed' X must be a square distance
            matrix. Otherwise it contains a sample per row. If the method
            is 'exact', X may be a sparse matrix of type 'csr', 'csc'
            or 'coo'.
        """

        X = check_array(X, dtype=np.float32, accept_sparse="csr", order="C")
        self._raw_data = X

        # Handle all the optional arguments, setting default
        if self.a is None or self.b is None:
            self.a, self.b = find_ab_params(self.spread, self.min_dist)

        self._validate_parameters()

        if self.verbose:
            print(str(self))

        index = list(range(X.shape[0]))

        # Error check n_neighbors based on data size
        if X[index].shape[0] <= self.n_neighbors:
            if X[index].shape[0] == 1:
                self.embedding_ = np.zeros(
                    (1, self.n_components)
                )  # needed to sklearn comparability
                return self

            warn(
                "n_neighbors is larger than the dataset size; truncating to "
                "X.shape[0] - 1"
            )
            self._n_neighbors = X[index].shape[0] - 1
        else:
            self._n_neighbors = self.n_neighbors

        # Note: unless it causes issues for setting 'index', could move this to
        # initial sparsity check above
        if self._sparse_data and not X.has_sorted_indices:
            X.sort_indices()

        random_state = check_random_state(self.random_state)

        if self.verbose:
            print(ts(), "Construct fuzzy simplicial set")

        # pass string identifier if pynndescent also defines distance metric
        if _HAVE_PYNNDESCENT:
            if self._sparse_data and self.metric in pynn_sparse_named_distances:
                nn_metric = self.metric
            elif not self._sparse_data and self.metric in pynn_named_distances:
                nn_metric = self.metric
            else:
                nn_metric = self._input_distance_func
        else:
            nn_metric = self._input_distance_func

        (self._knn_indices, self._knn_dists, _) = nearest_neighbors(
            X[index],
            self._n_neighbors,
            # int(self._n_neighbors * 1.2),  # we can use more neighbors
            nn_metric,
            self.angular_rp_forest,
            random_state,
            self.low_memory,
            use_pynndescent=True,
            verbose=self.verbose,
        )

        if self.local_n_epochs is None:
            self.local_n_epochs = 50

        if self.global_n_epochs is None:
            self.global_n_epochs = 100

        if self.verbose:
            print(ts(), "Build K-nearest neighbor graph structure")

        flat_indices = self._knn_indices.flatten()  # flattening all knn indices
        index, freq = np.unique(flat_indices, return_counts=True)
        # sorted_index = index[freq.argsort(kind="stable")]  # sorted index in increasing order
        sorted_index = index[
            freq.argsort(kind="stable")[::-1]
        ]  # sorted index in decreasing order

        # get disjoint NN matrix
        disjoints = build_knn_graph(
            data=X, sorted_index=sorted_index, hub_num=self.hub_num,
        )

        # get hub indices from disjoint set
        hubs = pick_hubs(disjoints=disjoints, random_state=random_state, popular=True,)

        if self.verbose:
            print(ts(), "Run global optimization")

        init_global = build_global_structure(
            data=X,
            hubs=hubs,
            n_components=self.n_components,
            a=self.a,
            b=self.b,
            random_state=random_state,
            alpha=self.global_learning_rate,
            n_epochs=self.global_n_epochs,
            verbose=self.verbose,
            label=self.ll,
            init_global=self.init,
        )

        if self.verbose:
            print(
                ts(), "Get NN indices & Initialize them using original hub information"
            )

        init, hub_info, hubs = embed_others_nn(
            data=X,
            init_global=init_global,
            hubs=hubs,
            knn_indices=self._knn_indices,
            nn_consider=self._n_neighbors,
            random_state=random_state,
            label=self.ll,
            verbose=self.verbose,
        )

        self._knn_indices, self._knn_dists, counts = select_from_knn(
            knn_indices=self._knn_indices,
            knn_dists=self._knn_dists,
            hub_info=hub_info,
            n_neighbors=self.n_neighbors,
            n=X.shape[0],
        )

        counts_hub = counts[hubs]
        counts_sum = len(counts_hub[counts_hub < self.n_neighbors])
        if counts_sum > 0:
            if self.verbose:
                print(ts(), "Adding more KNNs to build the graph")

            self._knn_indices, self._knn_dists, counts_sum = apppend_knn(
                data=X,
                knn_indices=self._knn_indices,
                knn_dists=self._knn_dists,
                hub_info=hub_info,
                n_neighbors=self.n_neighbors,
                counts=counts,
                counts_sum=counts_sum,
            )

            if counts_sum != 0:
                raise ValueError(
                    f"KNN indices not fully determined! counts_sum: {counts_sum} != 0"
                )

        self.graph_, _, _ = fuzzy_simplicial_set(
            X[hubs],
            self.n_neighbors,
            random_state,
            nn_metric,
            hubs,
            self._knn_indices[hubs],
            self._knn_dists[hubs],
            self.angular_rp_forest,
            self.set_op_mix_ratio,
            self.local_connectivity,
            True,
            True,
        )

        if self.verbose:
            print(ts(), "Run local optimization")

        init = local_optimize_nn(
            data=X,
            graph=self.graph_,
            hub_info=hub_info,
            n_components=self.n_components,
            learning_rate=self.local_learning_rate,
            a=self.a,
            b=self.b,
            gamma=self.gamma,
            negative_sample_rate=self.negative_sample_rate,
            n_epochs=self.local_n_epochs,
            init=init,
            random_state=random_state,
            parallel=False,
            verbose=self.verbose,
            label=self.ll,
        )

        if self.verbose:
            print(ts(), "Embedding outliers")

        self.embedding_ = embed_outliers(
            data=X,
            init=init,
            hubs=hubs,
            disjoints=disjoints,
            random_state=random_state,
            label=self.ll,
						n_neighbors=self.n_neighbors,
            verbose=self.verbose,
        )

        if self.verbose:
            print(ts(), "Finished embedding")

        self._input_hash = joblib.hash(self._raw_data)

        return self


    def fit_transform(self, X):
        """Fit X into an embedded space and return that transformed
        output.

        Parameters
        ----------
        X : array, shape (n_samples, n_features) or (n_samples, n_samples)
            If the metric is 'precomputed' X must be a square distance
            matrix. Otherwise it contains a sample per row.

        Returns
        -------
        X_new : array, shape (n_samples, n_components)
            Embedding of the training data in low-dimensional space.
        """
        self.fit(X)
        return self.embedding_



"""
Utils

This includes quantitative measures to compare how the high-dimensional data and its low embedding result are similar.

Provides
  0. Root Mean Squared Error (RMSE)
  1. Kruskal stress measure
  2. Sammon's stress
  3. Spearman's Rho
  4. Trustworthiness & Continuity
  5. Mean Relative Rank Error (MRRE)
  6. Distance To a Measure (DTM)
  7. KL Divergence between density distributions
  8. Set difference (TODO)
  9. Projection precision score (PPS) (TODO)
  10. Distance Consistency (TODO)
  11. ClustMe (TODO)

References
----------
.. [1] Kruskal's Stress Measure (1964)
    - J. Kruskal, Multidimensional scaling by optimizing goodness of fit to a nonmetric hypothesis, Psychometrika 29 (1964) 1-27.
    - J. Kruskal, Nonmetric multidimensional scaling: a numerical method, Psychometrika 29 (1964) 115-129. 

.. [2] Sammon's Stress (1969)
    - J.W. Sammon, A nonlinear mapping for data structure analysis, IEEE Trans. Comput. C-18 (1969). 
    - https://en.wikipedia.org/wiki/Sammon_mapping

.. [3] Spearman's Rho
    - Corder, G. W. & Foreman, D. I. (2014). Nonparametric Statistics: A Step-by-Step Approach, Wiley. ISBN 978-1118840313.
    - https://en.wikipedia.org/wiki/Spearman%27s_rank_correlation_coefficient

.. [4] Trustworthiness & Continuity (2001)
    - J. Venna, S. Kaski, Local multidimensional scaling, Neural Networks 19 (2006) 889-899.

.. [5] MRRE (2007)
    - J.A. Lee, M. Verleysen, Nonlinear dimensionality reduction, Springer, New York, London, 2007.
    - J.A. Lee, M. Verleysen, Rank-based quality assessment of nonlinear dimensionality reduction, in: ESANN, 2008, pp. 49-54.
    - J.A. Lee, M. Verleysen, Quality assessment of dimensionality reduction: rank-based criteria, Neurocomput 72 (2009) 1431-1443.
    - http://www.ecmlpkdd2008.org/files/pdf/workshops/fsdm/2.pdf 

.. [6] Distance to a measure (2011)
    - Frédéric Chazal, David Cohen-Steiner, and Quentin Mérigot. Geometric inference for probability measures.
    Foundations of Computational Mathematics, 11(6):733–751, 2011.
    - Chazal, F., Fasy, B., Lecci, F., Michel, B., Rinaldo, A., Rinaldo, A., & Wasserman, L. (2017).
    Robust topological inference: Distance to a measure and kernel distance. The Journal of Machine Learning Research, 18(1), 5845-5884.

.. [7] KL Divergence between density distributions (2020)
    - Moor, M., Horn, M., Rieck, B., & Borgwardt, K. (2020). Topological autoencoders. ICML.

.. [8] Set difference (2015) - focuses on neighborhood preservation
    - R. M. Martins, R. Minghim, and A. C. Telea, “Explaining Neighborhood Preservation for Multidimensional Projections,”
    in Proceedings of the Computer Graphics & Visual Computing (CGVC ’15). Eurographics, 2015, pp. 121–128.

.. [9] Projection precision score (2010) - local measure
    - T. Schreck, T. von Landesberger, and S. Bremm, “Techniques for Precision-Based Visual Analysis of Projected Data,”
    Information Visualization, vol. 9, no. 3, pp. 181–193, 2010.

.. [10] Distance consistency (2009)
    - M. Sips, B. Neubert, J. Lewis, and P. Hanrahan, “Selecting Good Views of High-Dimensional Data Using Class Consistency,”
    Computer Graphics Forum, vol. 28, no. 3, pp. 831–838, 2009.

.. [11] ClustMe (2019) - perception based measure
    - M. M. Abbas, M. Aupetit, M. Sedlmair, and H. Bensmail, “ClustMe: A Visual Quality Measure for Ranking Monochrome Scatterplots based on Cluster Patterns,”
    Computer Graphics Forum, vol. 38, no. 3, pp. 225–236, 2019.

.. [Survey]
    - L. G. Nonato and M. Aupetit, “Multidimensional Projection for Visual Analytics: Linking Techniques with Distortions, Tasks, and Layout Enrichment,”
    IEEE Transactions on Visualization and Computer Graphics, vol. 25, no. 8, pp. 2650–2673, 2019.
    - Gracia, A., González, S., Robles, V., & Menasalvas, E. (2014).
    A methodology to compare dimensionality reduction algorithms in terms of loss of quality. Information Sciences, 270, 1-27.

"""

import numpy as np
from scipy.stats import spearmanr
from sklearn import metrics
import numba
# import faiss


class GlobalMeasure:
    def __init__(self, x, z, dtype=np.float32):
        self.n_data = x.shape[0]  # number of data
        self.dtype = dtype

        # euclidean_distances: much faster than scipy.spatial.distance.squareform(pdist(x))
        # self.adjacency_matrix_z = metrics.pairwise.euclidean_distances(z)
        # self.adjacency_matrix_x = metrics.pairwise.euclidean_distances(x)

        """
        numba implementation is faster than above for large datasets
        e.g.) for KMNIST dataset
        calculating lower dimensional adjacency matrix: 57.7 (s) --> 34.2 (s)
        calculating high dimensional adjacency matrix: 684.0 (s) --> 71.2 (s)
        """

        self.adjacency_matrix_z = adjacency_matrix(data=z, dtype=self.dtype)
        self.adjacency_matrix_x = adjacency_matrix(data=x, dtype=self.dtype)

        if (self.adjacency_matrix_x.dtype != self.dtype) or (self.adjacency_matrix_z.dtype != self.dtype):
            self.adjacency_matrix_z = self.adjacency_matrix_z.astype(self.dtype)
            self.adjacency_matrix_x = self.adjacency_matrix_x.astype(self.dtype)

        # self.squared_differences = np.square(
        #     self.adjacency_matrix_x - self.adjacency_matrix_z, dtype=self.dtype
        # )

    # def rmse(self):
    #     """
    #     Root Mean Squared Error (RMSE)
    #     - Lower is BETTER
    #     - Global
    #     """
    #     sum_of_squared_differences = self.squared_differences.sum()

    #     return np.sqrt(sum_of_squared_differences / self.n_data, dtype=self.dtype)

    # def kruskal_stress_measure(self):
    #     """
    #     Kruskal Stress Measure

    #     A measure to capture the deviation from monotonicity
    #     - Lower is BETTER
    #     - Global
    #     """
    #     sum_of_squared_differences = self.squared_differences.sum()
    #     sum_of_squares_z = np.square(self.adjacency_matrix_z, dtype=self.dtype).sum()

    #     return np.sqrt(sum_of_squared_differences / sum_of_squares_z, dtype=self.dtype)

    # def sammon_stress(self):
    #     """
    #     Sammon's Stress

    #     An error measure used to test structure preservation
    #     - Lower is BETTER
    #     - Global
    #     """
    #     numerator = np.divide(
    #         self.squared_differences,
    #         self.adjacency_matrix_x,
    #         out=np.zeros(self.squared_differences.shape, dtype=float),
    #         where=self.adjacency_matrix_x != 0,
    #     )
    #     return numerator.sum() / self.adjacency_matrix_x.sum()

    def dtm(self, sigma=0.1):
        """
        Distance To a Measure (DTM)

        Compare normal distribution density between the original and embedding space
        - Lower is BETTER
        - Global

        Parameters
        ----------
        sigma : float
            sigma for normalization
        """
        density_x = self.dtm_calculation(self.adjacency_matrix_x, sigma=sigma)
        density_z = self.dtm_calculation(self.adjacency_matrix_z, sigma=sigma)
        return np.abs(density_x - density_z).sum()

    def dtm_kl(self, sigma=0.1):
        """
        KL Divergence between density distributions

        KL Divergence between normal distribution density of the original and embedding space
        - Lower is BETTER
        - Global

        Parameters
        ----------
        sigma : float
            sigma for normalization
        """
        density_x = self.dtm_calculation(self.adjacency_matrix_x, sigma=sigma)
        density_z = self.dtm_calculation(self.adjacency_matrix_z, sigma=sigma)
        density_kl = density_x * (np.log(density_x) - np.log(density_z))
        return density_kl.sum()

    @staticmethod
    def dtm_calculation(adjacency_matrix, sigma):
        # normalization using max value
        x = adjacency_matrix / adjacency_matrix.max()

        # get normalized density
        density_x = np.sum(np.exp(-(x ** 2) / sigma), axis=-1)
        return density_x / density_x.sum()


@numba.njit(
    locals={
        "matrix": numba.types.float32[:, ::1],
    },
    parallel=True,
    fastmath=True,
)
def adjacency_matrix(data, dtype):
    n = data.shape[0]
    dim = data.shape[1]
    matrix = np.zeros((n, n), dtype=dtype)

    for i in numba.prange(n):
        for j in numba.prange(i):
            for d in numba.prange(dim):
                matrix[i, j] += (data[i][d] - data[j][d]) ** 2

    matrix = matrix + matrix.T
    return np.sqrt(matrix)


@numba.njit(parallel=True)
def get_fast_knn(arr, n_neighbors):
    """
    requires much less memory && much less time for big data
    """
    knn_indices = np.empty((arr.shape[0], n_neighbors), dtype=np.int32)
    knn_ranks = np.empty((arr.shape[0], arr.shape[0]), dtype=np.int32)

    for i in numba.prange(arr.shape[0]):
        dists = np.empty(arr.shape[0], dtype=np.float32)
        for j in numba.prange(arr.shape[0]):
            dist = 0.0
            for d in numba.prange(arr.shape[1]):
                dist += (arr[i][d] - arr[j][d]) ** 2
            dists[j] = dist

        v = dists.argsort(kind="quicksort")
        knn_ranks[i] = v.argsort(kind="quicksort")
        knn_indices[i] = v[1 : n_neighbors + 1]

    return knn_indices, knn_ranks


# def get_gpu_knn(x, k):
#     index = faiss.IndexFlatL2(x.shape[1])
#     index.add(x)
#     _, knn_indices = index.search(x, k + 1) # first index = distance (we don't need it here)
#     # self.rank_x = 
#     return knn_indices[:, 1:k+1]


def get_nnidx_rank(arr, n_neighbors):
    """
    Brute force wau to get the index of NNs and ranks
    """
    adj_matrix = metrics.pairwise.euclidean_distances(arr)
    idx = adj_matrix.argsort()
    return idx[:, 1 : n_neighbors + 1], idx.argsort()


class LocalMeasure:
    def __init__(self, x, z, k=5):
        self.k = k  # number of nearest neighbors
        self.n_data = x.shape[0]  # number of data

        # slow && requires a lot of memory
        # self.nnidx_z, self.rank_z = get_nnidx_rank(z, k)
        # self.nnidx_x, self.rank_x = get_nnidx_rank(x, k)

        # fast && requires less memory
        self.nnidx_z, self.rank_z = get_fast_knn(z, k)
        self.nnidx_x, self.rank_x = get_fast_knn(x, k)

        # using GPU (faiss)
        # tmp = get_gpu_knn(x, k)


    def spearmans_rho(self):
        """
        Spearman's Rho

        This measure estimates the correlation of rank order data.
        It is defined as the Pearson correlation coefficient between the rank variables, 
        and can be viewed as one of the local neighborhood preservation measures.
        - Higher is BETTER (Preserved)
        - Local
        """

        ranks_list_x = []
        ranks_list_z = []
        for n in range(self.n_data):
            # get NNs in n-th row (in terms of high-dimensional space)
            rx = self.rank_x[n][self.nnidx_x[n]]
            rz = self.rank_z[n][self.nnidx_x[n]]
            # append them at once
            ranks_list_x.extend(rx)
            ranks_list_z.extend(rz)
        coeff, _ = spearmanr(ranks_list_x, ranks_list_z)

        return coeff

    def trustworthiness(self):
        """
        Trustworthiness

        A measure to check what extent the k nearest neighbours of a point are preserved when going from the original space to the latent space.
        If it is low it means that the data points originally farther away are captured as NNs of the embedding
        - Higher is BETTER (Preserved)
        - Local
        """
        return self._tc_calculation(
            self.nnidx_x, self.rank_x, self.nnidx_z, self.n_data, self.k
        )

    def continuity(self):
        """
        Continuity

        If it is low it means that data points that are originally close in high-dim are not captured as NNs in the embedding.
        - Higher is BETTER (Preserved)
        - Local
        """
        return self._tc_calculation(
            self.nnidx_z, self.rank_z, self.nnidx_x, self.n_data, self.k
        )

    @staticmethod
    def _tc_calculation(nnidx_base, rank, nnidx_target, n_data, k):
        value = 0.0

        # Calculate NNs comparing base and target space
        for n in range(n_data):
            # get NNs in lower dimension that was NOT NNs in high dimension
            missings = np.setdiff1d(nnidx_target[n], nnidx_base[n])

            for missing in missings:
                value += rank[n, missing] - k

        return 1 - 2 / (n_data * k * (2 * n_data - 3 * k - 1)) * value

    def mrre(self, ratio=1.0):
        """
        Mean Relative Rank Error (MRRE)

        Similar to trustworthiness & continuity but uses normalizing factor
        It denotes how well the NNs are preserved in terms of one space to another.
        - Higher is BETTER (Preserved)
        - Local

        Parameters
        ----------
        ratio : float
            ratio between 'mrre_zx' and 'mrre_xz',
            where mrre_zx ~= continuity and mrre_xz ~= trustworthiness
        """
        mrre_xz = self._mrre_caculation(
            self.nnidx_x, self.rank_x, self.rank_z, self.n_data, self.k
        )
        mrre_zx = self._mrre_caculation(
            self.nnidx_z, self.rank_z, self.rank_x, self.n_data, self.k
        )
        return mrre_xz * ratio + mrre_zx * (1 - ratio)

    def mrre_xz(self):
        mrre_xz = self._mrre_caculation(
            self.nnidx_x, self.rank_x, self.rank_z, self.n_data, self.k
        )
        return mrre_xz

    def mrre_zx(self):
        mrre_zx = self._mrre_caculation(
            self.nnidx_z, self.rank_z, self.rank_x, self.n_data, self.k
        )
        return mrre_zx

    @staticmethod
    def _mrre_caculation(nnidx_base, rank_base, rank_target, n_data, k):
        mrre_temp = 0.0
        for n in range(n_data):
            rank_targets = rank_target[n][nnidx_base[n]]
            rank_bases = rank_base[n][nnidx_base[n]]
            rank_norm = abs(rank_targets - rank_bases) / rank_bases
            mrre_temp += rank_norm.sum()

        # normalizing constant
        c = n_data * sum([abs(n_data - 2 * i + 1) / i for i in range(1, k + 1)])
        return 1 - mrre_temp / c



"""
This will compare embedding result quantitatively using functions in utils.py
"""

from .models.dataset import read_data
from .utils import GlobalMeasure, LocalMeasure
import pandas as pd
import argparse
import numpy as np


MEASURE_LOCAL_LIST = [
    "Spearman",
    "Trustworthiness",
    "Continuity",
    "MRRE",
]

ALGO_LIST = ["pca", "tsne", "umap", "topoae", "atsne", "umato"]
DATA_LIST = ["spheres"]


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="quantitative comparison of the embedding result"
    )
    parser.add_argument(
        "--data",
        type=str,
        help="choose dataset: spheres, mnist, fmnist, cifar10",
        default="spheres",
    )
    parser.add_argument(
        "--algo",
        type=str,
        help="choose algorithm: pca, tsne, umap, topoae, atsne, umato",
        default="all",
    )
    args = parser.parse_args()

    if args.algo != "all":
        ALGO_LIST = [args.algo]

    for k in [5]:
    # for k in [5, 10, 15, 20, 30, 35, 40, 45, 50, 75, 100]:
    # for k in [3,4,5,6,7,8,9,10,11,12,13]:
        measures = []
        algorithms = []
        values = []
        for alg in ALGO_LIST:

            print(f"[INFO] Test on [{args.data}] dataset using [{alg}] w/ {k} neighbors")

            # read data & embedding result
            x, z, label = read_data(args.data, alg)

            lmeasure = LocalMeasure(x, z, k=k)

            algorithms.extend([alg] * len(MEASURE_LOCAL_LIST))
            measures.extend(MEASURE_LOCAL_LIST)

            spearman_val = lmeasure.spearmans_rho()
            trust_val = lmeasure.trustworthiness()
            conti_val = lmeasure.continuity()
            mrre_val = lmeasure.mrre()

            values.extend(
                [spearman_val, trust_val, conti_val, mrre_val,]
                # [trust_val, conti_val, mrre_val,]
            )

            result = pd.DataFrame(
                {"measure": measures, "algorithm": algorithms, "values": values}
            )
            result = result.pivot(index="measure", columns="algorithm", values="values").fillna(
                "NA"
            )
        print(f"{result}\n")


from scipy.spatial import procrustes
import numpy as np
import pandas as pd
import os
import argparse

parser = argparse.ArgumentParser(description="args")
parser.add_argument("--data", type=str, help="choose dataset", required=True, default='spheres')

args = parser.parse_args()

if __name__ == "__main__":

    algos = ['tsne', 'umap', 'umato']
    # algos = ['umap', 'umato']

    for algo in algos:
        print(f"run algorithm: {algo}")
        path = os.path.join(os.getcwd(), "visualization", "public", "results", "init", args.data)

        d1= pd.read_csv(os.path.join(path, f'{algo}_{args.data}_class.csv'))
        d2= pd.read_csv(os.path.join(path, f'{algo}_{args.data}_random.csv'))
        d3= pd.read_csv(os.path.join(path, f'{algo}_{args.data}_spectral.csv'))
        d4= pd.read_csv(os.path.join(path, f'{algo}_{args.data}_pca.csv'))

        d1_no = d1.drop(columns=['label'])
        d1_no = np.array(d1_no)
        d2_no = d2.drop(columns=['label'])
        d2_no = np.array(d2_no)
        d3_no = d3.drop(columns=['label'])
        d3_no = np.array(d3_no)
        d4_no = d4.drop(columns=['label'])
        d4_no = np.array(d4_no)

        num = len(d1_no)
        ix = np.arange(num)

        ds = [d1_no, d2_no, d3_no, d4_no]
        disparities = []
        for i in enumerate(ds):
            for j in enumerate(ds):
                if i == j:
                    continue
                else:
                    _, _, disparity = procrustes(i[1], j[1])
                    disparities.append(disparity)
        print(f"mean value: {np.mean(disparities)}")
        print("====")